@@ -18,7 +18,8 @@ Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.*/
 #include <string.h>
 #include "header.h"
 #include "patchfunc.h"
-extern int indent;
+
+static int iterlev = 0;
 
 int has_iter(char *s)
 {
@@ -102,7 +103,10 @@ int guesstype(int n)
       return Gstr;
     }
   case Fentry:
-    return vartype(ctxstack[getvarerr(n)]);
+    {
+      long v = getvar(n);
+      return v<0 ? Ggen: vartype(ctxstack[v]);
+    }
   case Ffunction:
     if (!strcmp(entryname(n),"length"))
       return Glg;
@@ -166,8 +170,10 @@ int newgetg(int v, int l, const char *t)
 void makesubblock_ctx(int n, int savx)
 {
   int bl3=newblock();
+  int level = preclevel;
   genblock(n,-1);
   makeblock(bl3,n,newleaf(n),-1,savx);
+  preclevel = level;
 }
 
 void makesubblock(int n)
@@ -175,13 +181,15 @@ void makesubblock(int n)
   makesubblock_ctx(n, s_ctx.n);
 }
 
-Clabel_t genblock_label(int n, int p, int type)
+Clabel_t genblock_label(int n, int p, int type, int end)
 {
   Clabel_t t;
   static int labie=0;
   int l=newlabel(type);
   if (type==Liferr)
     label[l].ie=labie++;
+  if (type==Liter && end >=0)
+    label[l].iter=end;
   genblock(n,p);
   if (type==Liferr)
     labie--;
@@ -190,11 +198,11 @@ Clabel_t genblock_label(int n, int p, int type)
   return t;
 }
 
-Clabel_t makesubblock_label(int n, int t)
+Clabel_t makesubblock_label(int n, int t, int end)
 {
   int savx=s_ctx.n;
   int bl3=newblock();
-  Clabel_t lab=genblock_label(n,-1,t);
+  Clabel_t lab=genblock_label(n,-1,t,end);
   makeblock(bl3,n,newleaf(n),-1,savx);
   return lab;
 }
@@ -222,8 +230,31 @@ int get_iferr(int k)
   return 0;
 }
 
+int get_iter(int k)
+{
+  int i, l=s_label.n;
+  for (i=s_label.n-k; i<l; i++)
+  {
+    if (label[i].type==Liter)
+      return label[i].iter;
+  }
+  return 0;
+}
+
+int unwind_iter(int aseq, int k)
+{
+  int i, l=s_label.n-k;
+  for (i=s_label.n-1; i>=l; i--)
+  {
+    if (label[i].type==Liter)
+      aseq=addseqright(aseq,newleaf(label[i].iter));
+  }
+  return aseq;
+}
+
+
 int
-geniterproto(int n, long fl, gpfunc *gp, int *argo, int *seq, int *var, int *tv)
+geniterproto(int n, long fl, gpfunc *gp, int *argo, int *seq, int *var, int *tv, int *evar, int *etv)
 {
   int arg[STACKSZ];
   const char *p=gp->proto.code;
@@ -231,6 +262,7 @@ geniterproto(int n, long fl, gpfunc *gp, int *argo, int *seq, int *var, int *tv)
   PPproto mod;
   int nb=genlistargs(n,arg,0,STACKSZ);
   int i=0, k=0, binf;
+  *var = -1; *evar = -1;
   while((mod=parseproto(&p,&c))!=PPend)
   {
     if (i<nb && tree[arg[i]].f!=Fnoarg && (mod==PPdefault || mod==PPdefaultmulti))
@@ -253,6 +285,7 @@ geniterproto(int n, long fl, gpfunc *gp, int *argo, int *seq, int *var, int *tv)
         argo[k++] = binf;
         break;
       case 'V':
+        if (*var >= 0) { *evar = *var; *etv = *tv; }
         checkisvar(arg[i], gp->gpname, var, tv);
         if (fl)
         {
@@ -271,8 +304,12 @@ geniterproto(int n, long fl, gpfunc *gp, int *argo, int *seq, int *var, int *tv)
       break;
     case PPdefault:
     case PPdefaultmulti:
-      if (i<nb) tree[arg[i]].t = guesstype(arg[i]);
-      argo[k++] = i<nb ? arg[i] : GNOARG;
+      if (i < nb)
+      {
+        tree[arg[i]].t = guesstype(arg[i]);
+        if (c != 'I')
+          argo[k++] = arg[i];
+      }
       break;
     case PPstar:
       i=nb-1;
@@ -292,52 +329,95 @@ void
 genblockfunciter(int n, gpfunc *gp)
 {
   int arg[STACKSZ];
-  int var=-1, tv=Gnotype;
-  int bseq=-1,aseq=-1,ret=-1,vprime,seq,cseq;
+  int var, evar, vecvar=-1, tv=Gnotype, etv=Gnotype;
+  int bseq=-1,aseq=-1,ret=-1,vprime, seq=GNOARG, cseq;
   int gpm=-1,bl0,bl1,bl2;
   int sav0, savx;
   Clabel_t lab;
   gpiterator * iter;
-  const char * init, *next;
-  int nb = geniterproto(n, 1, gp, arg, &seq, &var, &tv);
+  const char * init, *next, *itername;
+  int nb = geniterproto(n, 1, gp, arg, &seq, &var, &tv, &evar, &etv);
   gpdescarg * da = descfindrulesdsc(nb, arg, gp->iter);
-  if (!da) die(n, "no suitable iterator found");
+  if (!da)
+  {
+    int i;
+    for(i=0;i<nb;i++)
+      fprintf(stderr,"%d: %s\n", i, GPname(tree[arg[i]].t));
+    die(n, "no suitable iterator found");
+  }
   iter = &(da->iter);
   init = lfunc[iter->init].gpname;
   next = lfunc[iter->next].gpname;
   sav0=s_ctx.n;
   bl0=newblock();
-  vprime=newnode(Fentry,newentry("iter"),-1);
+  if (iterlev==0)
+    itername = "iter";
+  else
+  {
+    char buf[64];
+    sprintf(buf,"iter%d",iterlev);
+    itername = strdup(buf);
+  }
+  iterlev++;
+  vprime=newnode(Fentry,newentry(itername),-1);
   pushvar(vprime,1<<Cuser,iter->type,-1);
-  nb = geniterproto(n, 0, gp, arg+1, &seq, &var, &tv)+1;
+  nb = geniterproto(n, 0, gp, arg+1, &seq, &var, &tv, &evar, &etv)+1;
   arg[0] = vprime;
   bseq = newcall(init, stacktoargs(arg,nb));
   genblock(bseq,-1);
   savx=s_ctx.n;
   bl1=newblock();
   bl2=newblock();
-  if (autogc)
+  if (seq!=GNOARG && autogc)
     gpm=gengerepile(bl2);
   if (tv==Gnotype)
   {
+    gpdescarg *dv;
     tree[vprime].t = iter->type;
-    gpdescarg *da=descfindrules(1, &vprime, lfunc+iter->next);
-    tv = da->type;
+    dv=descfindrules(1, &vprime, lfunc+iter->next);
+    if (!dv) die(n, "no suitable iterator_next found");
+    tv = dv->type;
+  }
+  if (evar>=0)
+  {
+    etv=tv; tv=Ggen;
+    pushvar(evar,(1<<Cuser)|(1<<Ccompo),etv,-1);
   }
   pushvar(var,1<<Cuser|1<<Ccompo,tv,-1);
-  lab=makesubblock_label(seq,Lbreak);
-  if (lab.ne)
-    seq=newseq(seq,newcall("_label",newsmall(lab.ne)));
-  if (autogc)
-    seq=addseqright(seq,newcall("_gerepilemany",gpm));
-  cseq=newnode(Fassign,var,newcall(next,vprime));
+  if (seq!=GNOARG)
+  {
+    if (iter->end>=0)
+    {
+      const char *end = lfunc[iter->end].gpname;
+      lab=makesubblock_label(seq, Liter, newcall(end,newleaf(vprime)));
+    } else
+      lab=makesubblock_label(seq,Lbreak,-1);
+  }
+  if (evar<0)
+    cseq = newnode1(Fassign,var,newcall(next,vprime));
+  else
+  {
+    newdecl((1<<Cauto)|(1<<Cconst),Gvec,-1,&vecvar);
+    cseq = newnode1(Fassign,vecvar,newcall(next,vprime));
+    seq = addseqleft(newnode1(Fassign,evar,newcoeff(vecvar,newsmall(1),-1)),seq);
+    seq = addseqleft(newnode1(Fassign,var,newcoeff(vecvar,newsmall(2),-1)),seq);
+  }
   tree[cseq].m=Mparens;
-  aseq=addseqright(aseq,newcall("_whilenull",newnode(Flistarg,cseq,seq)));
-  if (lab.go)
-    aseq=newseq(aseq,newcall("_label",newsmall(lab.go)));
+  if (seq!=GNOARG)
+  {
+    if (lab.ne)
+      seq=newseq(seq,newcall("_label",newsmall(lab.ne)));
+    if (autogc)
+      seq=addseqright(seq,newcall("_gerepilemany",gpm));
+    aseq=addseqright(aseq,newcall("_whilenull",newnode(Flistarg,cseq,seq)));
+    if (lab.go)
+      aseq=newseq(aseq,newcall("_label",newsmall(lab.go)));
+  } else
+    aseq=addseqright(aseq,newcall("_whilenull",cseq));
   makeblocks3(bl0,bl1,bl2,n,bseq,aseq,ret,savx,savx);
   copyctx(sav0,block+bl0);
   restorectx(sav0);
+  iterlev--;
 }
 
 /*
@@ -432,8 +512,11 @@ void genblockfuncspec(int n, int p, gpfunc *gp)
         arg[0] = newleaf(pred);
       }
       genblock(arg[0],n);
-      arg[1]=geninsertvar(arg[1],ret);
-      makesubblock(arg[1]);
+      if (arg[1]!=GNOARG)
+      {
+        arg[1]=geninsertvar(arg[1],ret);
+        makesubblock(arg[1]);
+      }
       if(nb==3)
       {
         arg[2]=geninsertvar(arg[2],ret==-1?-1:newleaf(ret));
@@ -503,7 +586,7 @@ void genblockfuncspec(int n, int p, gpfunc *gp)
       }
       if (tree[arg[1]].f==Ftag)
         functype(*gp)=tree[arg[1]].y;
-      for(;*p;p++) if (*p=='p') funcmode(*gp)|=(1<<Mprec);
+      for(;*p;p++) if (*p=='p' || *p=='b') funcmode(*gp)|=(1<<Mprec);
       gp->spec=GPinstalled;
     }
     break;
@@ -523,12 +606,12 @@ void genblockfuncspec(int n, int p, gpfunc *gp)
       gpm=gengerepile(bl2);
     genequaltosmall(binf,bsup,&tv,Ggen);
     pushvar(var,1<<Cuser,tv,-1);
-    lab=makesubblock_label(seq,Lbreak);
+    lab=makesubblock_label(seq,Lbreak,-1);
     if (lab.ne)
       seq=newseq(seq,newcall("_label",newsmall(lab.ne)));
     if (autogc)
       seq=addseqright(seq,newcall("_gerepilemany",gpm));
-    arg[0]=newnode(Fassign,var,vinf);
+    arg[0]=newnode1(Fassign,var,vinf);
     arg[1]=newopcall(OPle,newleaf(var),vsup);
     arg[2]=seq;
     arg[3]=newopcall(OPpp,newleaf(var),-1);
@@ -563,10 +646,10 @@ void genblockfuncspec(int n, int p, gpfunc *gp)
       bl2=newblock();
       if (autogc)
         gpm=gengerepile(bl2);
-      next = newnode(Fassign,newleaf(var),newnode(Ftag,
+      next = newnode1(Fassign,newleaf(var),newnode(Ftag,
             newcall("_call_GG",newnode(Flistarg,newnode(Flistarg,
                   newleaf(vfvnext),newleaf(vfvdat)),newleaf(var))),Gvec));
-      lab=makesubblock_label(seq,Lbreak);
+      lab=makesubblock_label(seq,Lbreak,-1);
       if (lab.ne)
         seq=newseq(seq,newcall("_label",newsmall(lab.ne)));
       if (autogc)
@@ -581,16 +664,17 @@ void genblockfuncspec(int n, int p, gpfunc *gp)
       makeblocks3(bl0,bl1,bl2,n,bseq,aseq,ret,sav0,savx);
     }
     break;
-  case GPfordiv:
+  case GPforeach:
     {
-      int vdiv,div;
-      int vloop;
+      int vdiv, vlen, vloop;
+      int vt = Ggen;
       nb=genlistargs(n,arg,3,3);
       seq=arg[2];
       checkisvar(arg[1],gp->gpname,&var,&tv);
-      newdecl(1<<Cconst,Gvec,-1,&vdiv);
-      div=newcall("divisors",arg[0]);
-      bseq=addseqright(bseq,geninsertvar(div,vdiv));
+      newdecl((1<<Cauto)|(1<<Cconst),Gempty,-1,&vdiv);
+      newdecl((1<<Cauto)|(1<<Cconst),Gempty,-1,&vlen);
+      bseq=addseqright(bseq,geninsertvar(arg[0],vdiv));
+      bseq=addseqright(bseq,geninsertvar(newcall("length",vdiv),vlen));
       genblock(bseq,-1);
       savx=s_ctx.n;
       bl1=newblock();
@@ -598,19 +682,62 @@ void genblockfuncspec(int n, int p, gpfunc *gp)
       if (autogc)
         gpm=gengerepile(bl2);
       newdecl(0,Gsmall,-1,&vloop);
-      if (tv==Gnotype) tv=Gint;
+      if (tv==Gnotype) tv = vt;
       pushvar(var,1<<Cuser,tv,-1);
-      lab=makesubblock_label(seq,Lbreak);
+      lab=makesubblock_label(seq,Lbreak,-1);
       if (lab.ne)
         seq=newseq(seq,newcall("_label",newsmall(lab.ne)));
       if (autogc)
         seq=addseqright(seq,newcall("_gerepilemany",gpm));
       seq=addseqleft(
-          newnode(Fassign,newleaf(var),
+          newnode1(Fassign,newleaf(var),
             newnode(Ftag,
-              newcoeff(newleaf(vdiv),newleaf(vloop),-1),Gint)),seq);
-      arg[0]=newnode(Fassign,vloop,newsmall(1));
-      arg[1]=newopcall(OPle,newleaf(vloop),newcall("length",vdiv));
+              newcoeff(newleaf(vdiv),newleaf(vloop),-1), vt)),seq);
+      arg[0]=newnode1(Fassign,vloop,newsmall(1));
+      arg[1]=newopcall(OPle,newleaf(vloop),newleaf(vlen));
+      arg[2]=seq;
+      arg[3]=newopcall(OPpp,newleaf(vloop),-1);
+      aseq=newnode(tree[n].f,newentry("for"),stacktoargs(arg,4));
+      if (lab.go)
+        aseq=newseq(aseq,newcall("_label",newsmall(lab.go)));
+      makeblocks(bl1,bl2,n,bseq,aseq,ret,savx);
+      break;
+    }
+  case GPfordivfactored:
+  case GPfordiv:
+    {
+      int vloop, vdiv, vlen, div;
+      int factored = gp->spec == GPfordivfactored;
+      int vt = factored ? Gvec: Gint;
+      nb=genlistargs(n,arg,3,3);
+      seq=arg[2];
+      checkisvar(arg[1],gp->gpname,&var,&tv);
+      newdecl((1<<Cauto)|(1<<Cconst),Gempty,-1,&vdiv);
+      newdecl((1<<Cauto)|(1<<Cconst),Gempty,-1,&vlen);
+      div=newcall("divisors", factored ?
+          newnode(Flistarg, arg[0], newsmall(1)) : arg[0]);
+      bseq=addseqright(bseq,geninsertvar(div,vdiv));
+      bseq=addseqright(bseq,geninsertvar(newcall("length",vdiv),vlen));
+      genblock(bseq,-1);
+      savx=s_ctx.n;
+      bl1=newblock();
+      bl2=newblock();
+      if (autogc)
+        gpm=gengerepile(bl2);
+      newdecl(0,Gsmall,-1,&vloop);
+      if (tv==Gnotype) tv = vt;
+      pushvar(var,1<<Cuser,tv,-1);
+      lab=makesubblock_label(seq,Lbreak,-1);
+      if (lab.ne)
+        seq=newseq(seq,newcall("_label",newsmall(lab.ne)));
+      if (autogc)
+        seq=addseqright(seq,newcall("_gerepilemany",gpm));
+      seq=addseqleft(
+          newnode1(Fassign,newleaf(var),
+            newnode(Ftag,
+              newcoeff(newleaf(vdiv),newleaf(vloop),-1), vt)),seq);
+      arg[0]=newnode1(Fassign,vloop,newsmall(1));
+      arg[1]=newopcall(OPle,newleaf(vloop),newleaf(vlen));
       arg[2]=seq;
       arg[3]=newopcall(OPpp,newleaf(vloop),-1);
       aseq=newnode(tree[n].f,newentry("for"),stacktoargs(arg,4));
@@ -640,10 +767,10 @@ void genblockfuncspec(int n, int p, gpfunc *gp)
         newdecl((1<<Cauto)|(1<<Cconst),Gsmall,-1,&vindex);
         newdecl((1<<Cauto)|(1<<Cconst),Gsmall,-1,&lstep);
         bseq=addseqright(bseq,geninsertvar(newcall("length",newleaf(vstep)),newleaf(lstep)));
-        sarg[0] = newnode(Fassign,newleaf(vindex),newsmall(1));
+        sarg[0] = newnode1(Fassign,newleaf(vindex),newsmall(1));
         sarg[1] = newleaf(lstep);
         sarg[2] = newnode(Fmatcoeff,newleaf(vstep),
-            newnode(Fmatrix,newnode(Frange,newleaf(vindex),GNOARG),-1));
+            newnode(Fmatrix,newnode(Frange,newleaf(vindex),GNORANGE),-1));
         sstep = newcall("sum", stacktoargs(sarg,3));
         if (tv==Gnotype) tv=Ggen;
       }
@@ -671,12 +798,12 @@ void genblockfuncspec(int n, int p, gpfunc *gp)
       if (autogc)
         gpm=gengerepile(bl2);
       pushvar(var,1<<Cuser,tv,-1);
-      lab=makesubblock_label(seq,Lbreak);
+      lab=makesubblock_label(seq,Lbreak,-1);
       if (lab.ne)
         seq=newseq(seq,newcall("_label",newsmall(lab.ne)));
       if (autogc)
         seq=addseqright(seq,newcall("_gerepilemany",gpm));
-      arg[0]=newnode(Fassign,var,vinf);
+      arg[0]=newnode1(Fassign,var,vinf);
       if ( !dosign )
       {
         if ( tree[bstep].x>0 )
@@ -745,7 +872,7 @@ void genblockfuncspec(int n, int p, gpfunc *gp)
     pushvar(var,1<<Cuser,tv,FC_forprime_init>=0?-1:newsmall(0));
     vprime=newnode(Fentry,newentry("primepointer"),-1);
     pushvar(vprime,1<<Cuser,Gforprime,FC_forprime_init>=0?-1:newcall("_diffptr",-1));
-    lab=makesubblock_label(seq,Lbreak);
+    lab=makesubblock_label(seq,Lbreak,-1);
     if (lab.ne)
       seq=newseq(seq,newcall("_label",newsmall(lab.ne)));
     if (autogc)
@@ -758,7 +885,7 @@ void genblockfuncspec(int n, int p, gpfunc *gp)
       if (bsup==GNOARG) nbs=2; else args[2]=vsup;
       aseq=addseqleft(newcall((tv==Gsmall?"_u_forprime_init":"_forprime_init"),
             stacktoargs(args,nbs)),aseq);
-      bseq=newnode(Fassign,var,newcall(tv==Gsmall?"_u_forprime_next":"_forprime_next_",vprime));
+      bseq=newnode1(Fassign,var,newcall(tv==Gsmall?"_u_forprime_next":"_forprime_next_",vprime));
       tree[bseq].m=Mparens;
       aseq=addseqright(aseq,newcall("_whilenull",newnode(Flistarg,bseq,seq)));
     }
@@ -812,10 +939,10 @@ void genblockfuncspec(int n, int p, gpfunc *gp)
     genequaltosmall(binf,bsup,&tv,Gint);
     pushvar(var,1<<Cuser,tv,-1);
     seq=geninsertvarop(seq,ret,(gp->spec==GPsum?OPpe:OPme));
-    makesubblock_label(seq,Lnobrk);
+    makesubblock_label(seq,Lnobrk,-1);
     if (autogc)
       seq=addseqright(seq,newcall("_gerepilemany",gpm));
-    arg[0]=newnode(Fassign,var,vinf);
+    arg[0]=newnode1(Fassign,var,vinf);
     arg[1]=newopcall(OPle,newleaf(var),vsup);
     arg[2]=seq;
     arg[3]=newopcall(OPpp,newleaf(var),-1);
@@ -844,14 +971,14 @@ void genblockfuncspec(int n, int p, gpfunc *gp)
       if (tv==Gnotype) tv=Gint;
       pushvar(var,1<<Cuser,tv,-1);
       seq=geninsertvarop(seq,ret,OPpe);
-      makesubblock_label(seq,Lnobrk);
+      makesubblock_label(seq,Lnobrk,-1);
       if (autogc)
         seq=addseqright(seq,newcall("_gerepilemany",gpm));
       seq=addseqleft(
-          newnode(Fassign,newleaf(var),
+          newnode1(Fassign,newleaf(var),
             newnode(Ftag,
               newcoeff(newleaf(vdiv),newleaf(vloop),-1),Gint)),seq);
-      arg[0]=newnode(Fassign,vloop,newsmall(1));
+      arg[0]=newnode1(Fassign,vloop,newsmall(1));
       arg[1]=newopcall(OPle,newleaf(vloop),newcall("length",vdiv));
       arg[2]=seq;
       arg[3]=newopcall(OPpp,newleaf(vloop),-1);
@@ -865,17 +992,22 @@ void genblockfuncspec(int n, int p, gpfunc *gp)
     bl1=newblock();
     savx=s_ctx.n;
     seq=arg[1];
-    if(autogc)
-      gpm=gengerepile(bl1);
-    genblock_label(arg[0],n,Lnobrk);
-    lab=makesubblock_label(seq,Lbreak);
-    if (lab.ne)
-      seq=newseq(seq,newcall("_label",newsmall(lab.ne)));
-    if (autogc)
-      seq=addseqright(seq,newcall("_gerepilemany",gpm));
+    if (seq != GNOARG)
+    {
+      if(autogc)
+        gpm=gengerepile(bl1);
+      genblock_label(arg[0],n,Lnobrk,-1);
+      lab=makesubblock_label(seq,Lbreak,-1);
+      if (lab.ne)
+        seq=newseq(seq,newcall("_label",newsmall(lab.ne)));
+      if (autogc)
+        seq=addseqright(seq,newcall("_gerepilemany",gpm));
+    }
+    else
+      genblock(arg[0],n);
     arg[1]=seq;
     aseq=newnode(tree[n].f,tree[n].x,stacktoargs(arg,nb));
-    if (lab.go)
+    if (seq != GNOARG && lab.go)
       aseq=newseq(aseq,newcall("_label",newsmall(lab.go)));
     makeblock(bl1,n,aseq,ret,savx);
     break;
@@ -900,11 +1032,11 @@ void genblockfuncspec(int n, int p, gpfunc *gp)
       newdecl(0,Gsmall,-1,&var);
     aseq=addseqright(aseq,newgetg(ret,newleaf(vsup),
           gp->spec==GPvector?"t_VEC":gp->spec==GPvectorv?"t_COL":"t_VECSMALL"));
-    arg[0]=newnode(Fassign,var,newsmall(1));
+    arg[0]=newnode1(Fassign,var,newsmall(1));
     arg[1]=newopcall(OPle,newleaf(var),vsup);
     arg[2]=newcoeff(newleaf(ret),newleaf(var),-1);
     arg[2]=geninsertvar(seq,arg[2]);
-    makesubblock_label(arg[2],Lnobrk);
+    makesubblock_label(arg[2],Lnobrk,-1);
     arg[3]=newopcall(OPpp,newleaf(var),-1);
     aseq=addseqright(aseq,newcall("for",stacktoargs(arg,4)));
     makeblocks(bl1,bl2,n,bseq,aseq,ret,savx);
@@ -944,15 +1076,15 @@ void genblockfuncspec(int n, int p, gpfunc *gp)
         seq=arg[4];
       aseq=newgetg(newcoeff(newleaf(ret),newleaf(var1),-1),
           newleaf(vsup2),"t_COL");
-      arg[0]=newnode(Fassign,var2,newsmall(1));
+      arg[0]=newnode1(Fassign,var2,newsmall(1));
       arg[1]=newopcall(OPle,newleaf(var2),vsup2);
       arg[3]=newopcall(OPpp,newleaf(var2),-1);
       arg[2]=newcoeff(newleaf(ret),newleaf(var2),newleaf(var1));
       arg[2]=geninsertvar(seq,arg[2]);
-      makesubblock_label(arg[2],Lnobrk);
+      makesubblock_label(arg[2],Lnobrk,-1);
       arg[2]=addseqright(aseq,newcall("for",stacktoargs(arg,4)));
       aseq=newgetg(ret,newleaf(vsup1),"t_MAT");
-      arg[0]=newnode(Fassign,var1,newsmall(1));
+      arg[0]=newnode1(Fassign,var1,newsmall(1));
       arg[1]=newopcall(OPle,newleaf(var1),vsup1);
       arg[3]=newopcall(OPpp,newleaf(var1),-1);
       aseq=addseqright(aseq,newcall("for",stacktoargs(arg,4)));
@@ -980,16 +1112,42 @@ void genblockfuncspec(int n, int p, gpfunc *gp)
       tree[n] = tree[aseq];
     }
     break;
+  case GPlocalprec:
+    {
+      char buf[64];
+      int vn;
+      sprintf(buf,"prec%d",++preclevel);
+      vn = newnode(Fentry, newentry(strdup(buf)), -1);
+      pushvar(vn, (1<<Cauto), Gsmall, -1);
+      y = newnode1(Fassign, vn, newcall("_ndec2prec", y));
+      genblock(y, n);
+      tree[n].y = y;
+    }
+    break;
+  case GPlocalbitprec:
+    {
+      char buf[64];
+      int vn;
+      sprintf(buf,"bitprec%d",++preclevel);
+      vn = newnode(Fentry, newentry(strdup(buf)), -1);
+      pushvar(vn, (1<<Cauto), Gsmall, -1);
+      y = newnode1(Fassign, vn, newnode(Ftag, newleaf(y), Gsmall));
+      genblock(y, n);
+      tree[n].y = y;
+    }
+    break;
   case GPreturn:
     {
-      int iferr_reset=0;
+      int iferr_reset=0, iter_reset=0;
       if (y==GNOARG || y<0)
         y=tree[n].y=GNIL;
       else
         genblock(y,n);
       iferr_reset=check_labels(s_label.n,Liferr)==1;
+      iter_reset=check_labels(s_label.n,Liter)==1;
+
       /* If we don't do garbage collecting, don't bother adding blocks.*/
-      if (!autogc && !iferr_reset) break;
+      if (!autogc && !iferr_reset && !iter_reset) break;
       if (tree[y].f==Fentry && !(ctxstack[getvarerr(y)].flag&(1<<Cglobal)))
       {
         var=y;
@@ -1022,6 +1180,8 @@ void genblockfuncspec(int n, int p, gpfunc *gp)
       }
       if (iferr_reset)
         aseq=addseqright(aseq,newcall("_iferr_CATCH_reset",newsmall(0)));
+      if (iter_reset)
+        aseq=unwind_iter(aseq,1);
       aseq=addseqright(aseq,newnode(tree[n].f,tree[n].x,y));
       makeblocks(bl1,bl2,n,bseq,aseq,ret,savx);
       break;
@@ -1065,6 +1225,11 @@ void genblockfuncspec(int n, int p, gpfunc *gp)
         aseq=addseqleft(newcall("_iferr_CATCH_reset",ie),newleaf(n));
         makeblock(newblock(),n,aseq,-1,s_ctx.n);
       }
+      if (check_labels(gp->spec==GPbreak?nb:nb-1,Liter))
+      {
+        aseq=addseqleft(unwind_iter(-1,nb),newleaf(n));
+        makeblock(newblock(),n,aseq,-1,s_ctx.n);
+      }
     }
     break;
   case GPiferr:
@@ -1083,7 +1248,7 @@ void genblockfuncspec(int n, int p, gpfunc *gp)
         newdecl((1<<Cauto)|(1<<Cimmutable),Gvoid,GNIL,&ret);
       seq=geninsertvar(seq,ret);
       bseq=geninsertvar(bseq,ret);
-      lab=makesubblock_label(seq,Liferr);
+      lab=makesubblock_label(seq,Liferr,-1);
       savx=s_ctx.n;
       pushvar(var,1<<Cuser,tv,newcall("_iferr_error",-1));
       if(nb==4)
@@ -1148,7 +1313,7 @@ int gentypefuncspec(int n, gpfunc *gp)
       tree[n].m|=(1<<Msidef);
     return tree[arg[1]].t;
   case GPwhilenull:
-    nb=genlistargs(n,arg,2,2);
+    nb=genlistargs(n,arg,1,2);
     gentypemode(nb,arg,&tree[n].m);
     tree[n].m&=~(1<<Mterm);
     return Gvoid;
@@ -1221,6 +1386,10 @@ int gentypefuncspec(int n, gpfunc *gp)
     gentype(y);
     tree[n].m=(tree[y].m&MODHERIT)|(1<<Msidef);
     return Gvec;
+  case GPmakevecsmall:
+    gentype(y);
+    tree[n].m=(tree[y].m&MODHERIT)|(1<<Msidef);
+    return Gvecsmall;
   case GPclosure:
     nb=genlistargs(n,arg,1,STACKSZ);
     gentypemode(nb,arg,&tree[n].m);
@@ -1248,7 +1417,7 @@ void gencopyarg(FILE *fout)
   long i;
   if (tree[n].x!=ufunc->bl || !fc->var) return;
   for (i=0;i<fc->s.n;i++)
-    if (fc->c[i].flag&(1<<Carg))
+    if (fc->c[i].flag&(1<<Carg) && !(fc->c[i].flag&(1<<Cindarg)))
     {
       ctxvar *v=fc->c+i;
       long j;
@@ -1327,11 +1496,16 @@ void genentryspec(FILE *fout, int n, gpfunc *gp)
       fprintf(fout,"\n");
       break;
     case GPwhilenull:
-      nb=genlistargs(n,arg,2,2);
+      nb=genlistargs(n,arg,1,2);
       genindentline(fout,"while (");
       genparens(fout,arg[0]);
-      fprintf(fout,")\n");
-      genbrace(fout,arg[1]);
+      if (nb==1)
+        fprintf(fout,");\n");
+      else
+      {
+        fprintf(fout,")\n");
+        genbrace(fout,arg[1]);
+      }
       break;
     case GPwhile:
       nb=genlistargs(n,arg,2,2);
@@ -1339,8 +1513,13 @@ void genentryspec(FILE *fout, int n, gpfunc *gp)
       {
         genindentline(fout,"while (");
         gencast(fout,arg[0],Gbool);
-        fprintf(fout,")\n");
-        genbrace(fout,arg[1]);
+        if (tree[arg[1]].f==Fnoarg)
+          fprintf(fout,");\n");
+        else
+        {
+          fprintf(fout,")\n");
+          genbrace(fout,arg[1]);
+        }
       }
       else
       {
@@ -1379,9 +1558,14 @@ void genentryspec(FILE *fout, int n, gpfunc *gp)
       nb=genlistargs(n,arg,2,2);
       if ( tree[arg[0]].f!=Fseq && tree[arg[0]].f!=Fblock)
       {
-        genindentline(fout,"do\n");
-        genbrace(fout,arg[1]);
-        fprintf(fout," while(");/*KB style*/
+        if (arg[1]!=GNOARG)
+        {
+          genindentline(fout,"do\n");
+          genbrace(fout,arg[1]);
+          fprintf(fout," while(");/*KB style*/
+        }
+        else
+          genindentline(fout,"while(");
         gencast(fout,arg[0],Gnegbool);
         fprintf(fout,");\n");
       }
@@ -1494,6 +1678,7 @@ void genentryspec(FILE *fout, int n, gpfunc *gp)
     fprintf(fout,")\n");
     genbrace(fout,arg[1]);
     break;
+  case GPmakevecsmall:
   case GPmakevec:
     nb=genlistargs(n,arg,1,STACKSZ);
     genindent(fout);
@@ -1558,6 +1743,11 @@ void genentryspec(FILE *fout, int n, gpfunc *gp)
       fprintf(fout,";\n");
     }
     break;
+  case GPlocalprec:
+  case GPlocalbitprec:
+    gencode(fout, y);
+    preclevel++;
+    break;
   case GPreturn:
     fprintf(fout,"return");
     if (functype(lfunc[currfunc])!=Gvoid)
